open Bigarray

type int_array = (int, int_elt, c_layout) Array1.t 
type byte_array = (int, int8_unsigned_elt, c_layout) Array1.t

let cmp s i j =
  let rec loop n (s : byte_array) i j =
    if i < n && j < n then begin
      match s.{i} - s.{j} with
          0 -> loop n s (i + 1) (j + 1)
        | n -> n
    end else if i = n then 1 else -1
  in loop (Array1.dim s) s i j

let swap (a: int_array) i j =
  let t = a.{i} in
  a.{i} <- a.{j};
  a.{j} <- t

let invoke (f : 'a -> 'b) x : unit -> 'b =
  let input, output = Unix.pipe() in
  match Unix.fork() with
    | -1 -> Unix.close input; Unix.close output; (let v = f x in fun () -> v)
    | 0 ->
      Unix.close input;
	let output = Unix.out_channel_of_descr output in
        Marshal.to_channel output (try `Res(f x) with e -> `Exn e) [];
        close_out output;
        exit 0
  | pid ->
      Unix.close output;
      let input = Unix.in_channel_of_descr input in fun () ->
        let v = Marshal.from_channel input in
        ignore (Unix.waitpid [] pid);
          close_in input;
        match v with `Res x -> x | `Exn e -> raise e

let invoke_serial f x () = f x

let sort str (a: int_array) =
  let rec qsort l u =
    if l < u then
      begin
        swap a l ((l + u) / 2);
        let m = ref l in
        for i=l+1 to u do
          if cmp str a.{i} a.{l} < 0 then
            begin
              incr m;
              swap a !m i
            end
        done;
        swap a l !m;
        let f = (if u - l > 200000 then invoke else invoke_serial)
                  (qsort l) (!m - 1)
        in
          qsort (!m + 1) u;
          f ()
      end
  in qsort 0 (Array1.dim a - 1)

let () =
  let file = try Array.get Sys.argv 1 with _ -> "input.txt" in
  let desc = Unix.openfile file [] 777 in
  let str = Array1.map_file desc int8_unsigned c_layout false (-1) in
  let n = Array1.dim str in
  let a = 
        Array1.map_file (Unix.openfile "/dev/zero" [Unix.O_RDWR] 777)
          int c_layout true n in
  for i=0 to n-1 do a.{i} <- i done;
  sort str a;
  for i=0 to n-1 do
    print_char(Char.chr str.{(a.{i} + n - 1) mod n})
  done;

  Unix.close desc
